#!/usr/bin/env python3
#
# -- General configuration ------------------------------------------------
from sphinx_automodapi import __version__

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autosummary']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = '.rst'

# The main toctree document.
master_doc = 'index'

# General information about the project.
project = 'sphinx-automodapi'
copyright = '2016, The Astropy Developers'
author = 'The Astropy Developers'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags.
release = __version__
# The short X.Y version.
version = '.'.join(release.split('.')[:2])

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_rtd_theme'

# Output file base name for HTML help builder.
htmlhelp_basename = 'sphinx-automodapidoc'

nitpicky = True



import os
import sys
import subprocess as sp

from textwrap import dedent

import pytest


@pytest.fixture
def cython_testpackage(tmpdir, request):
    """
    Creates a trivial Cython package for use with tests.
    """

    test_pkg = tmpdir.mkdir('test_pkg')
    test_pkg.mkdir('apyhtest_eva').ensure('__init__.py')
    test_pkg.join('apyhtest_eva').join('unit02.pyx').write(dedent("""\
        def pilot():
            \"\"\"Returns the pilot of Eva Unit-02.\"\"\"

            return True
    """))

    import sphinx_automodapi  # noqa

    test_pkg.join('setup.py').write(dedent("""\
        import sys

        sys.path.insert(0, {!r})

        from os.path import join
        from setuptools import setup, Extension

        NAME = 'apyhtest_eva'
        VERSION = 0.1
        RELEASE = True


        setup(
            name=NAME,
            version=VERSION,
            ext_modules=[Extension('apyhtest_eva.unit02',
                                   [join('apyhtest_eva', 'unit02.pyx')])]
        )
    """.format(os.path.dirname(sphinx_automodapi.__path__[0]))))

    # Build the Cython module in a subprocess; otherwise strange things can
    # happen with Cython's global module state
    sp.call([sys.executable, 'setup.py', 'build_ext', '--inplace'], cwd=test_pkg.strpath)

    sys.path.insert(0, str(test_pkg))
    import apyhtest_eva.unit02  # noqa

    def cleanup(test_pkg=test_pkg):
        for modname in ['apyhtest_eva', 'apyhtest_eva.unit02']:
            try:
                del sys.modules[modname]
            except KeyError:
                pass

        sys.path.remove(str(test_pkg))

    request.addfinalizer(cleanup)

    return test_pkg



# The following tests use a plain Python example module that is at
# sphinx_automodapi.tests.example_module.

# We store different cases in the cases sub-directory of the tests directory

import os
import glob
import shutil
from itertools import product

import pytest

from copy import deepcopy, copy
from sphinx.util.osutil import ensuredir
from docutils.parsers.rst import directives, roles

from .helpers import build_main, write_conf

CASES_ROOT = os.path.join(os.path.dirname(__file__), 'cases')

CASES_DIRS = glob.glob(os.path.join(CASES_ROOT, '*'))

PARALLEL = {False, True}
PROP_ATTR = {True, False}


intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None)
}

DEFAULT_CONF = {'source_suffix': '.rst',
                'master_doc': 'index',
                'nitpicky': True,
                'extensions': ['sphinx.ext.intersphinx', 'sphinx_automodapi.automodapi'],
                'suppress_warnings': ['app.add_directive', 'app.add_node'],
                'intersphinx_mapping': intersphinx_mapping,
                'nitpick_ignore': [('py:class', 'sphinx_automodapi.tests.example_module.classes.BaseSpam'),
                                   ('py:class', 'sphinx_automodapi.tests.example_module.other_classes.BaseFoo'),
                                   # See the following links for why these classes need to be ignored.
                                   # This only seems to be necessary for Python 2.7.
                                   #
                                   # https://trac.sagemath.org/ticket/19211
                                   # https://stackoverflow.com/q/11417221/3776794
                                   ('py:class', '_abcoll.Sequence'),
                                   ('py:class', '_abcoll.Iterable'),
                                   ('py:class', '_abcoll.Container'),
                                   ('py:class', '_abcoll.Sized')]}


def setup_function(func):
    # This can be replaced with the docutils_namespace context manager once
    # it is in a stable release of Sphinx
    func._directives = copy(directives._directives)
    func._roles = copy(roles._roles)


def teardown_function(func):
    directives._directives = func._directives
    roles._roles = func._roles


@pytest.mark.parametrize(('case_dir', 'parallel', 'prop_attr'), product(CASES_DIRS, PARALLEL, PROP_ATTR))
def test_run_full_case(tmpdir, case_dir, parallel, prop_attr):

    input_dir = os.path.join(case_dir, 'input')

    output_folder = "output_prop_is_attr" if prop_attr else "output_prop_is_not_attr"
    output_dir = os.path.join(case_dir, output_folder)
    if not os.path.isdir(output_dir):
        output_dir = os.path.join(case_dir, 'output')

    docs_dir = tmpdir.mkdir('docs').strpath

    conf = deepcopy(DEFAULT_CONF)
    conf.update({'automodapi_toctreedirnm': 'api',
                 'automodapi_writereprocessed': True,
                 'automodsumm_writereprocessed': True,
                 'automodsumm_properties_are_attributes': prop_attr})

    if os.path.basename(case_dir) in ('mixed_toplevel',
                                      'mixed_toplevel_all_objects',
                                      'allowed_names'):
        conf['extensions'].append('sphinx_automodapi.smart_resolver')

    start_dir = os.path.abspath('.')

    src_dir = 'src' if 'source_dir' in case_dir else '.'

    ensuredir(os.path.join(docs_dir, src_dir))

    write_conf(os.path.join(os.path.join(docs_dir, src_dir), 'conf.py'), conf)

    for root, dirnames, filenames in os.walk(input_dir):
        for filename in filenames:
            root_dir = os.path.join(docs_dir, os.path.relpath(root, input_dir))
            ensuredir(root_dir)
            input_file = os.path.join(root, filename)
            shutil.copy(input_file, root_dir)

    argv = ['-W', '-b', 'html', src_dir, '_build/html']
    if parallel:
        argv.insert(0, '-j 4')

    try:
        os.chdir(docs_dir)
        status = build_main(argv=argv)
    finally:
        os.chdir(start_dir)

    assert status == 0

    # Check that all expected output files are there and match the reference files
    for root, dirnames, filenames in os.walk(output_dir):
        for filename in filenames:
            path_reference = os.path.join(root, filename)
            path_relative = os.path.relpath(path_reference, output_dir)
            path_actual = os.path.join(docs_dir, path_relative)
            assert os.path.exists(path_actual)
            with open(path_actual, encoding='utf8') as f:
                actual = f.read()
            with open(path_reference, encoding='utf8') as f:
                reference = f.read()
            assert actual.strip() == reference.strip()


def test_duplicated_warning(tmpdir):
    input_dir = os.path.join(os.path.dirname(__file__), 'duplicated_warning', 'docs')
    docs_dir = tmpdir.mkdir('docs').strpath

    start_dir = os.path.abspath('.')
    src_dir = '.'

    for root, dirnames, filenames in os.walk(input_dir):
        for filename in filenames:
            root_dir = os.path.join(docs_dir, os.path.relpath(root, input_dir))
            ensuredir(root_dir)
            input_file = os.path.join(root, filename)
            shutil.copy(input_file, root_dir)

    argv = ['-W', '-b', 'html', src_dir, '_build/html']

    try:
        os.chdir(docs_dir)
        status = build_main(argv=argv)
    finally:
        os.chdir(start_dir)

    assert status == 0


def test_slots_example():
    """Basic tests for slots example module"""
    from sphinx_automodapi.tests.example_module.slots import (
        SlotDict, DerivedParam, DerivedSlotParam
    )
    SlotDict('param', 'other_param').my_method()
    DerivedParam('param', 'other_param', 'extra_param').derived_from_slot_class_method()
    DerivedSlotParam('param', 'other_param', 'extra_param').derived_from_slot_class_method()



import sys
import re
import os
from inspect import ismodule
from warnings import warn

import sphinx
from packaging.version import Version
from sphinx.ext.autosummary.generate import find_autosummary_in_docstring

__all__ = ['cleanup_whitespace',
           'find_mod_objs',
           'find_autosummary_in_lines_for_automodsumm']

SPHINX_LT_9 = Version(sphinx.__version__) < Version("9.0")

# We use \n instead of os.linesep because even on Windows, the generated files
# use \n as the newline character.
SPACE_NEWLINE = ' \n'
SINGLE_NEWLINE = '\n'
DOUBLE_NEWLINE = '\n\n'
TRIPLE_NEWLINE = '\n\n\n'


def cleanup_whitespace(text):
    """
    Make sure there are never more than two consecutive newlines, and that
    there are no trailing whitespaces.
    """

    # Get rid of overall leading/trailing whitespace
    text = text.strip() + '\n'

    # Get rid of trailing whitespace on each line
    while SPACE_NEWLINE in text:
        text = text.replace(SPACE_NEWLINE, SINGLE_NEWLINE)

    # Avoid too many consecutive newlines
    while TRIPLE_NEWLINE in text:
        text = text.replace(TRIPLE_NEWLINE, DOUBLE_NEWLINE)

    return text


def find_mod_objs(modname, onlylocals=False, sort=False):
    """ Returns all the public attributes of a module referenced by name.

    .. note::
        The returned list *not* include subpackages or modules of
        `modname`,nor does it include private attributes (those that
        beginwith '_' or are not in `__all__`).

    Parameters
    ----------
    modname : str
        The name of the module to search.
    onlylocals : bool or list
        If `True`, only attributes that are either members of `modname` OR one of
        its modules or subpackages will be included.  If a list, only members
        of packages in the list are included. If `False`, selection is done.
        This option is ignored if a module defines __all__ - in that case, __all__
        is used to determine whether objects are public.

    Returns
    -------
    localnames : list of str
        A list of the names of the attributes as they are named in the
        module `modname` .
    fqnames : list of str
        A list of the full qualified names of the attributes (e.g.,
        ``astropy.utils.misc.find_mod_objs``). For attributes that are
        simple variables, this is based on the local name, but for
        functions or classes it can be different if they are actually
        defined elsewhere and just referenced in `modname`.
    objs : list of objects
        A list of the actual attributes themselves (in the same order as
        the other arguments)

    """

    __import__(modname)
    mod = sys.modules[modname]

    # Note: use getattr instead of mod.__dict__[k] for modules that
    # define their own __getattr__ and __dir__.
    if hasattr(mod, '__all__'):
        pkgitems = [(k, getattr(mod, k)) for k in mod.__all__]
        # Optionally sort the items alphabetically
        if sort:
            pkgitems.sort()
        onlylocals = False
    else:
        pkgitems = [(k, getattr(mod, k)) for k in dir(mod) if k[0] != "_"]

    # filter out modules and pull the names and objs out
    localnames = [k for k, v in pkgitems if not ismodule(v)]
    objs = [v for k, v in pkgitems if not ismodule(v)]

    # fully qualified names can be determined from the object's module
    fqnames = []
    for obj, lnm in zip(objs, localnames):
        if hasattr(obj, '__module__') and hasattr(obj, '__name__'):
            fqnames.append(obj.__module__ + '.' + obj.__name__)
        else:
            fqnames.append(modname + '.' + lnm)

    if onlylocals:
        if isinstance(onlylocals, (tuple, list)):
            modname = tuple(onlylocals)
        valids = [fqn.startswith(modname) for fqn in fqnames]
        localnames = [e for i, e in enumerate(localnames) if valids[i]]
        fqnames = [e for i, e in enumerate(fqnames) if valids[i]]
        objs = [e for i, e in enumerate(objs) if valids[i]]

    return localnames, fqnames, objs


def find_autosummary_in_lines_for_automodsumm(lines, module=None, filename=None):
    """Find out what items appear in autosummary:: directives in the
    given lines.
    Returns a list of (name, toctree, template, inherited_members, noindex)
    where *name* is a name
    of an object and *toctree* the :toctree: path of the corresponding
    autosummary directive (relative to the root of the file name),
    *template* the value of the :template: option, and *inherited_members*
    is the value of the :inherited-members: option.
    *toctree*, *template*, and *inherited_members* are ``None`` if the
    directive does not have the corresponding options set.

    .. note::

       This is a slightly modified version of
       ``sphinx.ext.autosummary.generate.find_autosummary_in_lines``
       which recognizes the ``inherited-members`` option.
    """
    autosummary_re = re.compile(r'^(\s*)\.\.\s+autosummary::\s*')
    automodule_re = re.compile(
        r'^\s*\.\.\s+automodule::\s*([A-Za-zäüöÄÜÖßő0-9_.]+)\s*$')
    module_re = re.compile(
        r'^\s*\.\.\s+(current)?module::\s*([a-zA-ZäüöÄÜÖßő0-9_.]+)\s*$')
    autosummary_item_re = re.compile(r'^\s+(~?[_a-zA-ZäüöÄÜÖßő][a-zA-ZäüöÄÜÖßő0-9_.]*)\s*.*?')
    toctree_arg_re = re.compile(r'^\s+:toctree:\s*(.*?)\s*$')
    template_arg_re = re.compile(r'^\s+:template:\s*(.*?)\s*$')
    inherited_members_arg_re = re.compile(r'^\s+:inherited-members:\s*$')
    no_inherited_members_arg_re = re.compile(r'^\s+:no-inherited-members:\s*$')
    noindex_arg_re = re.compile(r'^\s+:noindex:\s*$')
    other_options_re = re.compile(r'^\s+:nosignatures:\s*$')

    documented = []

    toctree = None
    template = None
    inherited_members = None
    noindex = None
    current_module = module
    in_autosummary = False
    base_indent = ""

    for line in lines:
        if in_autosummary:
            m = toctree_arg_re.match(line)
            if m:
                toctree = m.group(1)
                if filename:
                    toctree = os.path.join(os.path.dirname(filename),
                                           toctree)
                continue

            m = template_arg_re.match(line)
            if m:
                template = m.group(1).strip()
                continue

            m = inherited_members_arg_re.match(line)
            if m:
                inherited_members = True
                continue

            m = no_inherited_members_arg_re.match(line)
            if m:
                inherited_members = False
                continue

            m = noindex_arg_re.match(line)
            if m:
                noindex = True
                continue

            if line.strip().startswith(':'):
                if other_options_re.match(line):
                    continue  # skip known options
                else:
                    warn(line)  # warn about unknown options

            m = autosummary_item_re.match(line)
            if m:
                name = m.group(1).strip()
                if name.startswith('~'):
                    name = name[1:]
                if current_module and \
                   not name.startswith(current_module + '.'):
                    name = f"{current_module}.{name}"
                documented.append((name, toctree, template,
                                   inherited_members, noindex))
                continue

            if not line.strip() or line.startswith(base_indent + " "):
                continue

            in_autosummary = False

        m = autosummary_re.match(line)
        if m:
            in_autosummary = True
            base_indent = m.group(1)
            toctree = None
            template = None
            inherited_members = None
            continue

        m = automodule_re.search(line)
        if m:
            current_module = m.group(1).strip()
            # recurse into the automodule docstring
            documented.extend(find_autosummary_in_docstring(
                current_module, filename=filename))
            continue

        m = module_re.match(line)
        if m:
            current_module = m.group(2)
            continue

    return documented


# sphinx-automodapi used to use sphinx.ext.autosummary.get_documenter()
# from Sphinx proper, but the function was removed upstream in
# https://github.com/sphinx-doc/sphinx/pull/13985.
def get_object_type(app, obj, parent):
    """Get the object type suitable for documenting the given object.

    *obj* is the Python object to be documented, and *parent* is an
    another Python object (e.g. a module or a class) to which *obj*
    belongs to.
    """
    if SPHINX_LT_9:
        from sphinx.ext.autosummary import get_documenter

        documenter = get_documenter(app, obj, parent)
        obj_type = documenter.objtype
        return obj_type

    from sphinx.ext.autosummary import _get_documenter

    obj_type = _get_documenter(obj, parent)
    return obj_type



from collections.abc import Sequence

__all__ = ['SequenceSubclass']


class SequenceSubclass(Sequence):
    """
    Inherits from an ABC.
    """

    def __init__(self):
        self._items = []

    def __len__(self):
        """
        Must be defined for the collections.abc.Sequence base class.
        """
        return len(self._items)

    def __getitem__(self, key):
        """
        Must be defined for the collections.abc.Sequence base class.
        """
        return self._items[key]

    def my_method(self, parameter):
        """
        An example method.
        """
        pass

    @property
    def my_property(self):
        """
        An example property.
        """
        return 42



"""
A collection of useful variables.
"""

__all__ = ['PARROT_STATE', 'FUNNY_WALK_STEPS']


PARROT_STATE = 'dead'
"""The global state of the parrot."""

FUNNY_WALK_STEPS = [['left', 'right'],
                    ['left', 'jump', 'right', 'jump'],
                    ['swim']]
"""List of different possible walk.

Each item contains a list of steps.
"""

# A variable not in __all__ should not be propagated.
NOTHING_HAPPENS = 0

# Even if it has a docstring
REALLY_NOTHING = 1
"""Really nothing."""



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from copy import copy

import pytest
from docutils.parsers.rst import directives, roles

from . import cython_testpackage  # noqa
from .helpers import run_sphinx_in_tmpdir


def setup_function(func):
    # This can be replaced with the docutils_namespace context manager once
    # it is in a stable release of Sphinx
    func._directives = copy(directives._directives)
    func._roles = copy(roles._roles)


def teardown_function(func):
    directives._directives = func._directives
    roles._roles = func._roles


# nosignatures

ADD_RST = """
:orphan:

add
===

.. currentmodule:: sphinx_automodapi.tests.example_module.mixed

.. autofunction:: add
""".strip()

MIXEDSPAM_RST = """
:orphan:

MixedSpam
=========

.. currentmodule:: sphinx_automodapi.tests.example_module.mixed

.. autoclass:: MixedSpam
   :show-inheritance:
""".strip()


def write_api_files_to_tmpdir(tmpdir):
    apidir = tmpdir.mkdir('api')
    with open(apidir.join('sphinx_automodapi.tests.example_module.mixed.add.rst').strpath, 'w') as f:
        f.write(ADD_RST)
    with open(apidir.join('sphinx_automodapi.tests.example_module.mixed.MixedSpam.rst').strpath, 'w') as f:
        f.write(MIXEDSPAM_RST)


ams_to_asmry_str = """
Before

.. automodsumm:: sphinx_automodapi.tests.example_module.mixed
{options}

And After
"""

ams_to_asmry_expected = """\
.. currentmodule:: sphinx_automodapi.tests.example_module.mixed

.. autosummary::

    add
    MixedSpam

"""


def test_ams_to_asmry(tmpdir):

    with open(tmpdir.join('index.rst').strpath, 'w') as f:
        f.write(ams_to_asmry_str.format(options=''))

    write_api_files_to_tmpdir(tmpdir)

    run_sphinx_in_tmpdir(tmpdir)

    with open(tmpdir.join('index.rst.automodsumm').strpath) as f:
        result = f.read()

    assert result == ams_to_asmry_expected


def test_too_many_options(tmpdir, capsys):

    ops = ['', ':classes-only:', ':functions-only:']
    ostr = '\n    '.join(ops)

    with open(tmpdir.join('index.rst').strpath, 'w') as f:
        f.write(ams_to_asmry_str.format(options=ostr))

    write_api_files_to_tmpdir(tmpdir)

    run_sphinx_in_tmpdir(tmpdir, expect_error=True)

    stdout, stderr = capsys.readouterr()
    assert ("[automodsumm] Defined more than one of functions-only, "
            "classes-only, and variables-only.  Skipping this directive." in stderr)


ORDEREDDICT_RST = """
:orphan:

OrderedDict
===========

.. currentmodule:: sphinx_automodapi.tests.example_module.noall

.. autoclass:: OrderedDict
   :show-inheritance:
""".strip()


@pytest.mark.parametrize('options,expect', [
    ('', ['add', 'MixedSpam']),
    (':allowed-package-names: sphinx_automodapi', ['add', 'MixedSpam']),
    (':allowed-package-names: collections', ['OrderedDict']),
    (':allowed-package-names: sphinx_automodapi,collections',
     ['add', 'MixedSpam', 'OrderedDict']),
])
def test_am_allowed_package_names(options, expect, tmpdir):
    """
    Test that allowed_package_names is interpreted correctly.
    """
    def mixed2noall(s):
        return s.replace('example_module.mixed', 'example_module.noall')

    am_str = ams_to_asmry_str
    with open(tmpdir.join('index.rst').strpath, 'w') as f:
        f.write(mixed2noall(am_str).format(options=('   '+options if options else '')))

    apidir = tmpdir.mkdir('api')
    with open(apidir.join('sphinx_automodapi.tests.example_module.noall.add.rst').strpath, 'w') as f:
        f.write(mixed2noall(ADD_RST))
    with open(apidir.join('sphinx_automodapi.tests.example_module.noall.MixedSpam.rst').strpath, 'w') as f:
        f.write(mixed2noall(MIXEDSPAM_RST))
    with open(apidir.join('sphinx_automodapi.tests.example_module.noall.OrderedDict.rst').strpath, 'w') as f:
        f.write(ORDEREDDICT_RST)

    run_sphinx_in_tmpdir(tmpdir)

    with open(tmpdir.join('index.rst.automodsumm').strpath) as f:
        result = f.read()

    for x in expect:
        assert '   '+x in result


PILOT_RST = """
:orphan:

pilot
=====

.. currentmodule:: apyhtest_eva.unit02

.. autofunction:: pilot
""".strip()

ams_cython_str = """
Before

.. automodsumm:: apyhtest_eva.unit02
    :functions-only:

And After
"""

ams_cython_expected = """\
.. currentmodule:: apyhtest_eva.unit02

.. autosummary::

    pilot

"""


def test_ams_cython(tmpdir, cython_testpackage):  # noqa

    with open(tmpdir.join('index.rst').strpath, 'w') as f:
        f.write(ams_cython_str)

    apidir = tmpdir.mkdir('api')
    with open(apidir.join('apyhtest_eva.unit02.pilot.rst').strpath, 'w') as f:
        f.write(PILOT_RST)

    run_sphinx_in_tmpdir(tmpdir)

    with open(tmpdir.join('index.rst.automodsumm').strpath) as f:
        result = f.read()

    assert result == ams_cython_expected


# =============================================================================

CLASS_RST = """
:orphan:

.. currentmodule:: {mod}

.. autoclass:: {cls}
""".strip()

sorted_str = """
Before

.. automodsumm:: sphinx_automodapi.tests.example_module.classes
    :sort:

And After
"""

sorted_expected = """\
.. currentmodule:: sphinx_automodapi.tests.example_module.classes

.. autosummary::

    Egg
    Spam

"""


def test_sort(tmpdir):
    with open(tmpdir.join("index.rst").strpath, "w") as f:
        f.write(sorted_str)

    apidir = tmpdir.mkdir('api')
    mod = 'sphinx_automodapi.tests.example_module.classes'
    for cls in "Spam", "Egg":
        with open(apidir.join(f'{mod}.{cls}.rst').strpath, 'w') as f:
            f.write(CLASS_RST.format(mod=mod, cls=cls))

    run_sphinx_in_tmpdir(tmpdir)

    with open(tmpdir.join("index.rst.automodsumm").strpath) as f:
        result = f.read()

    assert result == sorted_expected
